import pytest
from spharpy.samplings import Coordinates
import pyfar as pf
import numpy as np


@pytest.fixture
def make_coordinates():

    class Factory:
        @staticmethod
        def create_coordinates(
                implementation='spharpy', rad=1, theta=np.pi/2, phi=np.pi/2):

            if implementation == 'pyfar':
                return pf.Coordinates(
                    phi, theta, rad, domain='sph', convention='top_colat'
                )
            elif implementation == 'spharpy':
                return Coordinates.from_spherical(rad, theta, phi)

    yield Factory


@pytest.fixture
def icosahedron():
    """Return the coordinate points of an icosahedron in spherical coordinates

    Returns
    -------
    rad : float, ndarray
        The radius in meters.
    theta : float, ndarray
        The colatitude angle in radians.
    phi : float, ndarray
        The azimuth angle in radians.
    """
    gamma_R_r = np.arccos(np.cos(np.pi/3) / np.sin(np.pi/5))
    gamma_R_rho = np.arccos(1/(np.tan(np.pi/5) * np.tan(np.pi/3)))

    theta = np.tile(np.array([
        np.pi - gamma_R_rho,
        np.pi - gamma_R_rho - 2*gamma_R_r,
        2*gamma_R_r + gamma_R_rho,
        gamma_R_rho]), 5)
    theta = np.sort(theta)
    phi = np.arange(0, 2*np.pi, 2*np.pi/5)
    phi = np.concatenate((np.tile(phi, 2), np.tile(phi + np.pi/5, 2)))
    rad = np.ones(20)

    return rad, theta, phi



"""
Tests for spherical harmonic basis and related functions
"""
import spharpy.spherical as sh
import numpy as np
import numpy.testing as npt
import pytest


@pytest.mark.parametrize("implementation", ['spharpy', 'pyfar'])
def test_spherical_harmonic(make_coordinates, implementation):
    Nmax = 1
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)
    rad = np.ones(3, dtype=float)

    coords = make_coordinates.create_coordinates(
        implementation, rad, theta, phi)

    Y = np.array([[2.820947917738781e-01 + 0.000000000000000e+00j, 3.454941494713355e-01 + 0.000000000000000e+00j, 2.991827511286337e-17 + 0.000000000000000e+00j, -3.454941494713355e-01 + 0.000000000000000e+00j],
                  [2.820947917738781e-01 + 0.000000000000000e+00j, 2.115541521371041e-17 - 3.454941494713355e-01j, 2.991827511286337e-17 + 0.000000000000000e+00j, -2.115541521371041e-17 - 3.454941494713355e-01j],
                  [2.820947917738781e-01 + 0.000000000000000e+00j, 0.000000000000000e+00 + 0.000000000000000e+00j, 4.886025119029199e-01 + 0.000000000000000e+00j, 0.000000000000000e+00 + 0.000000000000000e+00j]], dtype=complex)

    basis = sh.spherical_harmonic_basis(Nmax, coords)

    np.testing.assert_allclose(Y, basis, atol=1e-13)


@pytest.mark.parametrize("implementation", ['spharpy', 'pyfar'])
def test_spherical_harmonic_n10(make_coordinates, implementation):
    Nmax = 10
    theta = np.array([np.pi/2, np.pi/2, 0], dtype=float)
    phi = np.array([0, np.pi/2, 0], dtype=float)

    coords = make_coordinates.create_coordinates(
        implementation, np.ones_like(theta), theta, phi)

    Y = np.genfromtxt(
        './tests/data/sh_basis_cplx_n10.csv',
        delimiter=',',
        dtype=complex)

    basis = sh.spherical_harmonic_basis(Nmax, coords)

    np.testing.assert_allclose(Y, basis, atol=1e-13)


@pytest.mark.parametrize("implementation", ['spharpy', 'pyfar'])
def test_spherical_harmonics_real(make_coordinates, implementation):
    n_max = 10
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2], dtype=float)
    phi = np.array([0, np.pi/2, 0, np.pi/4], dtype=float)
    rad = np.ones(4)

    coords = make_coordinates.create_coordinates(
        implementation, rad, theta, phi)

    reference = np.genfromtxt('./tests/data/sh_basis_real.csv', delimiter=',')
    basis = sh.spherical_harmonic_basis_real(n_max, coords)
    np.testing.assert_allclose(basis, reference, atol=1e-13)


@pytest.mark.parametrize("implementation", ['spharpy', 'pyfar'])
def test_orthogonality(make_coordinates, implementation):
    """
    Check if the orthonormality condition of the spherical harmonics is
    fulfilled
    """
    n_max = 82
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2], dtype=float)
    phi = np.array([0, np.pi/2, 0, np.pi/4], dtype=float)
    n_points = phi.size
    n_points = len(theta)
    coords = make_coordinates.create_coordinates(
        implementation, np.ones_like(theta), theta, phi)
    basis = sh.spherical_harmonic_basis(n_max, coords)

    inner = (basis @ np.conjugate(basis.T))
    fact = 4 * np.pi / (n_max + 1) ** 2
    orth = np.diagonal(fact * inner)
    np.testing.assert_allclose(
        orth, np.ones(n_points, dtype=complex), rtol=1e-10)


@pytest.mark.parametrize("implementation", ['spharpy', 'pyfar'])
def test_orthogonality_real(make_coordinates, implementation):
    """
    Check if the orthonormality condition of the reavl valued spherical
    harmonics is fulfilled
    """
    n_max = 82
    theta = np.array([np.pi / 2, np.pi / 2, 0, np.pi / 2], dtype='double')
    phi = np.array([0, np.pi / 2, 0, np.pi / 4], dtype='double')
    n_points = phi.size

    coords = make_coordinates.create_coordinates(
        implementation, np.ones_like(theta), theta, phi)

    basis = sh.spherical_harmonic_basis_real(n_max, coords)

    inner = (basis @ np.conjugate(basis.T))
    fact = 4 * np.pi / (n_max + 1) ** 2
    orth = np.diagonal(fact * inner)
    np.testing.assert_allclose(orth, np.ones(n_points), rtol=1e-10, atol=1e-10)


@pytest.mark.parametrize("implementation", ['spharpy', 'pyfar'])
def test_spherical_harmonic_basis_gradient(make_coordinates, implementation):
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    coords = make_coordinates.create_coordinates(
        implementation, np.ones_like(theta), theta, phi)

    grad_ele, grad_azi = \
        sh.spherical_harmonic_basis_gradient(n_max, coords)

    desire_ele = np.genfromtxt(
        './tests/data/Y_grad_ele.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_ele, desire_ele, rtol=1e-10, atol=1e-10)

    desire_azi = np.genfromtxt(
        './tests/data/Y_grad_azi.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_azi, desire_azi, rtol=1e-10, atol=1e-10)


@pytest.mark.parametrize("implementation", ['spharpy', 'pyfar'])
def test_spherical_harmonic_basis_gradient_real(
        make_coordinates, implementation):
    n_max = 15
    theta = np.array([np.pi/2, np.pi/2, 0, np.pi/2, np.pi/4])
    phi = np.array([0, np.pi/2, 0, np.pi/4, np.pi/4])

    coords = make_coordinates.create_coordinates(
        implementation, np.ones_like(theta), theta, phi)

    grad_ele, grad_azi = \
        sh.spherical_harmonic_basis_gradient_real(n_max, coords)

    desire_ele = np.genfromtxt(
        './tests/data/Y_grad_real_ele.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_ele, desire_ele, rtol=1e-10, atol=1e-10)

    desire_azi = np.genfromtxt(
        './tests/data/Y_grad_real_azi.csv',
        dtype=complex,
        delimiter=',')
    npt.assert_allclose(grad_azi, desire_azi, rtol=1e-10, atol=1e-10)



from spharpy.samplings import Coordinates
from spharpy.samplings import SamplingSphere
import pyfar as pf


def convert_coordinates(coordinates):
    coords_type = type(coordinates)
    if coords_type is not pf.Coordinates:
        return coordinates
    if coordinates.sh_order is None:
        return Coordinates.from_pyfar(coordinates)
    else:
        return SamplingSphere.from_pyfar(coordinates)



# -*- coding: utf-8 -*-

"""Top-level package for spharpy."""

__author__ = """The pyfar developers"""
__email__ = 'info@pyfar.org'
__version__ = '0.6.2'

from . import spherical
from . import samplings
from . import plot
from . import indexing
from . import transforms
from . import beamforming
from . import interpolate
from . import _deprecation


__all__ = [
    'spherical',
    'samplings',
    'plot',
    'indexing',
    'transforms',
    'beamforming',
    'interpolate',
    '_deprecation',
]
